@@ -14,17 +14,24 @@
 -- You should have received a copy of the GNU Affero General Public License
 -- along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-{-# LANGUAGE CPP, FlexibleContexts, ConstraintKinds #-}
+{-# LANGUAGE CPP, RecordWildCards, FlexibleContexts, ConstraintKinds #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving, DeriveDataTypeable, DeriveGeneric #-}
 
 module Distribution.Helper (
     Programs(..)
+  , defaultPrograms
+  , QueryEnv
+  , qeReadProcess
+  , qePrograms
+  , qeProjectDir
+  , qeDistDir
+  , qeCabalPkgDb
+  , qeCabalVer
+  , defaultQueryEnv
 
   -- * Running Queries
   , Query
   , runQuery
-  , runQuery'
-  , runQuery''
 
   -- * Queries against Cabal\'s on disk state
 
@@ -36,6 +43,11 @@ module Distribution.Helper (
   , ghcPkgOptions
   , ghcMergedPkgOptions
   , ghcLangOptions
+  , pkgLicenses
+  , flags
+  , configFlags
+  , nonDefaultConfigFlags
+  , packageId
 
   -- * Result types
   , ChModuleName(..)
@@ -67,7 +79,6 @@ import Control.Monad.Reader
 import Control.Exception as E
 import Data.Char
 import Data.List
-import Data.Default
 import Data.Version
 import Data.Typeable
 import Distribution.Simple.BuildPaths (exeExtension)
@@ -81,7 +92,7 @@ import GHC.Generics
 import Prelude
 
 import Paths_cabal_helper (getLibexecDir)
-import CabalHelper.Types
+import CabalHelper.Types hiding (Options(..))
 import CabalHelper.Sandbox
 
 -- | Paths or names of various programs we need.
@@ -91,8 +102,45 @@ data Programs = Programs {
       ghcPkgProgram :: FilePath
     } deriving (Eq, Ord, Show, Read, Generic, Typeable)
 
-instance Default Programs where
-    def = Programs "cabal" "ghc" "ghc-pkg"
+defaultPrograms :: Programs
+defaultPrograms = Programs "cabal" "ghc" "ghc-pkg"
+
+data QueryEnv = QueryEnv {
+      -- | How to start the cabal-helper process. Useful if you need to
+      -- capture stderr output from the helper.
+      qeReadProcess :: FilePath -> [String] -> String -> IO String,
+
+      qePrograms    :: Programs,
+
+      -- | Path to project directory, i.e. the one containing the
+      -- @project.cabal@ file
+      qeProjectDir  :: FilePath,
+
+      -- | Path to the @dist/@ directory
+      qeDistDir     :: FilePath,
+
+      -- | Where to look for the Cabal library when linking the helper
+      qeCabalPkgDb  :: Maybe FilePath,
+
+      -- | If @dist/setup-config@ wasn\'t written by this version of Cabal throw
+      -- an error
+      qeCabalVer    :: Maybe Version
+    }
+
+defaultQueryEnv :: FilePath
+                -- ^ Path to project directory, i.e. the one containing the
+                -- @project.cabal@ file
+                -> FilePath
+                -- ^ Path to the @dist/@ directory
+                -> QueryEnv
+defaultQueryEnv projdir distdir = QueryEnv {
+    qeReadProcess = readProcess
+  , qePrograms    = defaultPrograms
+  , qeProjectDir  = projdir
+  , qeDistDir     = distdir
+  , qeCabalPkgDb  = Nothing
+  , qeCabalVer    = Nothing
+  }
 
 data SomeLocalBuildInfo = SomeLocalBuildInfo {
       slbiPackageDbStack      :: [ChPkgDb],
@@ -102,7 +150,11 @@ data SomeLocalBuildInfo = SomeLocalBuildInfo {
       slbiGhcSrcOptions       :: [(ChComponentName, [String])],
       slbiGhcPkgOptions       :: [(ChComponentName, [String])],
       slbiGhcMergedPkgOptions :: [String],
-      slbiGhcLangOptions      :: [(ChComponentName, [String])]
+      slbiGhcLangOptions      :: [(ChComponentName, [String])],
+      slbiPkgLicenses         :: [(String, [(String, Version)])],
+      slbiFlags               :: [(String, Bool)],
+      slbiConfigFlags         :: [(String, Bool)],
+      slbiNonDefaultConfigFlags :: [(String, Bool)]
     } deriving (Eq, Ord, Read, Show)
 
 -- | Caches helper executable result so it doesn't have to be run more than once
@@ -112,54 +164,20 @@ newtype Query m a = Query { unQuery :: StateT (Maybe SomeLocalBuildInfo)
                                          (ReaderT QueryEnv m) a }
     deriving (Functor, Applicative, Monad, MonadIO)
 
-data QueryEnv = QueryEnv {
-      _qeReadProcess :: FilePath -> [String] -> String -> IO String,
-      _qeProgs       :: Programs,
-      _qeProjectDir  :: FilePath,
-      _qeDistDir     :: FilePath
-    }
-
 type MonadQuery m = ( MonadIO m
                     , MonadState (Maybe SomeLocalBuildInfo) m
                     , MonadReader QueryEnv m)
 
-run :: Monad m
-  => QueryEnv -> Maybe SomeLocalBuildInfo -> Query m a -> m a
+run :: Monad m => QueryEnv -> Maybe SomeLocalBuildInfo -> Query m a -> m a
 run e s action = flip runReaderT e (flip evalStateT s (unQuery action))
 
 -- | @runQuery query distdir@. Run a 'Query'. @distdir@ is where Cabal's
 -- @setup-config@ file is located.
 runQuery :: Monad m
-         => FilePath -- ^ Path to project directory, i.e. the one containing the
-                     -- @project.cabal@ file
-         -> FilePath -- ^ Path to @dist/@
+         => QueryEnv
          -> Query m a
          -> m a
-runQuery pd dd action = run (QueryEnv readProcess def pd dd) Nothing action
-
-runQuery' :: Monad m
-         => Programs
-         -> FilePath -- ^ Path to project directory, i.e. the one containing the
-                     -- @project.cabal@ file
-         -> FilePath -- ^ Path to @dist/@
-         -> Query m a
-         -> m a
-runQuery' progs pd dd action =
-    run (QueryEnv readProcess progs pd dd) Nothing action
-
-runQuery'' :: Monad m
-         => (FilePath -> [String] -> String -> IO String)
-         -- ^ How to start the cabal-helper process. Useful if you need to
-         -- capture stderr output from the helper.
-         -> Programs
-         -> FilePath -- ^ Path to project directory, i.e. the one containing the
-                     -- @project.cabal@ file
-         -> FilePath -- ^ Path to @dist/@
-         -> Query m a
-         -> m a
-runQuery'' readProc progs pd dd action =
-    run (QueryEnv readProc progs pd dd) Nothing action
-
+runQuery qe action = run qe Nothing action
 
 getSlbi :: MonadQuery m => m SomeLocalBuildInfo
 getSlbi = do
@@ -198,6 +216,24 @@ ghcMergedPkgOptions :: MonadIO m => Query m [String]
 -- | Only language related options, i.e. @-XSomeExtension@
 ghcLangOptions :: MonadIO m => Query m [(ChComponentName, [String])]
 
+-- | Get the licenses of the packages the current project is linking against.
+pkgLicenses :: MonadIO m => Query m [(String, [(String, Version)])]
+
+-- | Flag definitions from cabal file
+flags :: MonadIO m => Query m [(String, Bool)]
+
+-- | Flag assignments from setup-config
+configFlags :: MonadIO m => Query m [(String, Bool)]
+
+-- | Flag assignments from setup-config which differ from the default
+-- setting. This can also include flags which cabal decided to modify,
+-- i.e. don't rely on these being the flags set by the user directly.
+nonDefaultConfigFlags :: MonadIO m => Query m [(String, Bool)]
+
+-- | Package identifier, i.e. package name and version
+packageId :: MonadIO m => Query m (String, Version)
+
+
 packageDbStack      = Query $ slbiPackageDbStack      `liftM` getSlbi
 entrypoints         = Query $ slbiEntrypoints         `liftM` getSlbi
 sourceDirs          = Query $ slbiSourceDirs          `liftM` getSlbi
@@ -206,6 +242,11 @@ ghcSrcOptions       = Query $ slbiGhcSrcOptions       `liftM` getSlbi
 ghcPkgOptions       = Query $ slbiGhcPkgOptions       `liftM` getSlbi
 ghcMergedPkgOptions = Query $ slbiGhcMergedPkgOptions `liftM` getSlbi
 ghcLangOptions      = Query $ slbiGhcLangOptions      `liftM` getSlbi
+pkgLicenses         = Query $ slbiPkgLicenses         `liftM` getSlbi
+flags               = Query $ slbiFlags               `liftM` getSlbi
+configFlags         = Query $ slbiConfigFlags         `liftM` getSlbi
+nonDefaultConfigFlags = Query $ slbiNonDefaultConfigFlags `liftM` getSlbi
+packageId           = Query $ getPackageId
 
 -- | Run @cabal configure@
 reconfigure :: MonadIO m
@@ -218,38 +259,57 @@ reconfigure readProc progs cabalOpts = do
             [ "--with-ghc=" ++ ghcProgram progs ]
             -- Only pass ghc-pkg if it was actually set otherwise we
             -- might break cabal's guessing logic
-            ++ if ghcPkgProgram progs /= ghcPkgProgram def
+            ++ if ghcPkgProgram progs /= "ghc-pkg"
                  then [ "--with-ghc-pkg=" ++ ghcPkgProgram progs ]
                  else []
             ++ cabalOpts
     _ <- liftIO $ readProc (cabalProgram progs) ("configure":progOpts) ""
     return ()
 
-getSomeConfigState :: MonadQuery m => m SomeLocalBuildInfo
-getSomeConfigState = ask >>= \(QueryEnv readProc progs projdir distdir) -> do
-  let progArgs = [ "--with-ghc="     ++ ghcProgram progs
+
+
+invokeHelper :: MonadQuery m => [String] -> m [Maybe ChResponse]
+invokeHelper args = ask >>= \QueryEnv {..} -> do
+  let progs = qePrograms
+      projdir = qeProjectDir
+      distdir = qeDistDir
+
+      progArgs = [ "--with-ghc="     ++ ghcProgram progs
                  , "--with-ghc-pkg=" ++ ghcPkgProgram progs
                  , "--with-cabal="   ++ cabalProgram progs
                  ]
 
-  let args = [ "package-db-stack"
-             , "entrypoints"
-             , "source-dirs"
-             , "ghc-options"
-             , "ghc-src-options"
-             , "ghc-pkg-options"
-             , "ghc-merged-pkg-options"
-             , "ghc-lang-options"
-             ] ++ progArgs
-
-  res <- liftIO $ do
+  liftIO $ do
     exe  <- findLibexecExe "cabal-helper-wrapper"
-    out <- readProc exe (projdir:distdir:args) ""
+    let args' = progArgs ++ projdir:distdir:args
+    out <- qeReadProcess exe args' ""
     evaluate (read out) `E.catch` \(SomeException _) ->
-      error $ concat ["getSomeConfigState", ": ", exe, " "
-                     , intercalate " " (map show $ distdir:args)
-                     , " (read failed)"]
+      error $ concat ["invokeHelper", ": ", exe, " "
+                     , intercalate " " (map show args')
+                     , " (read failed)"
+                     ]
 
+getPackageId :: MonadQuery m => m (String, Version)
+getPackageId = ask >>= \QueryEnv {..} -> do
+  [ Just (ChResponseVersion pkgName pkgVer) ] <- invokeHelper [ "package-id" ]
+  return (pkgName, pkgVer)
+
+
+getSomeConfigState :: MonadQuery m => m SomeLocalBuildInfo
+getSomeConfigState = ask >>= \QueryEnv {..} -> do
+  res <- invokeHelper [ "package-db-stack"
+                      , "entrypoints"
+                      , "source-dirs"
+                      , "ghc-options"
+                      , "ghc-src-options"
+                      , "ghc-pkg-options"
+                      , "ghc-merged-pkg-options"
+                      , "ghc-lang-options"
+                      , "licenses"
+                      , "flags"
+                      , "config-flags"
+                      , "non-default-config-flags"
+                      ]
   let [ Just (ChResponsePkgDbs pkgDbs),
         Just (ChResponseEntrypoints eps),
         Just (ChResponseCompList srcDirs),
@@ -257,10 +317,15 @@ getSomeConfigState = ask >>= \(QueryEnv readProc progs projdir distdir) -> do
         Just (ChResponseCompList ghcSrcOpts),
         Just (ChResponseCompList ghcPkgOpts),
         Just (ChResponseList     ghcMergedPkgOpts),
-        Just (ChResponseCompList ghcLangOpts) ] = res
+        Just (ChResponseCompList ghcLangOpts),
+        Just (ChResponseLicenses pkgLics),
+        Just (ChResponseFlags fls),
+        Just (ChResponseFlags cfls),
+        Just (ChResponseFlags ndcfls)
+        ] = res
 
   return $ SomeLocalBuildInfo
-    pkgDbs eps srcDirs ghcOpts ghcSrcOpts ghcPkgOpts ghcMergedPkgOpts ghcLangOpts
+    pkgDbs eps srcDirs ghcOpts ghcSrcOpts ghcPkgOpts ghcMergedPkgOpts ghcLangOpts pkgLics fls cfls ndcfls
 
 -- | Make sure the appropriate helper executable for the given project is
 -- installed and ready to run queries.
